#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This script is an excerpt of the code used in the paper:

    Caron, S., Pham, Q. C., & Nakamura, Y. (2014). "Kinodynamic Motion Retiming
    for Humanoid Robots".  Proceedings of the 32nd Annual Conference of the
    Robotics Society of Japan (RSJ), Fukuoka, Japan.

"""

import cvxopt.solvers

from constraint import TOPPConstraint
from numpy import array, dot, zeros, linalg, allclose, eye

cvxopt.solvers.options['show_progress'] = False  # disable cvxopt output


class ForceConstraint(TOPPConstraint):

    def compute_abc(self, q, qs, qss):
        """
        Compute vectors a(s), b(s), c(s) for use with TOPP.

        q -- joint-angle values
        qs -- joint-angle velocities
        qss -- joint-angle accelerations

        """
        JT = self.robot.compute_contact_jacobian(q, qs).T
        cdim = JT.shape[1]  # == 3 * nb force vectors
        pinvJP = linalg.pinv(JT[50:, :])
        nullspace_proj = eye(cdim) - dot(pinvJP, JT[50:, :])

        with self.robot.rave:
            self.robot.rave.SetDOFValues(q)
            self.robot.rave.SetDOFVelocities(qs)
            tMqss, tCqs, tg = self.robot.rave.ComputeInverseDynamics(
                qss, externalforcetorque=None, returncomponents=True)
            tMqs, _, _ = self.robot.rave.ComputeInverseDynamics(
                qs, externalforcetorque=None, returncomponents=True)

        act_proj = lambda v: v[:50]
        cf_proj = lambda v: dot(pinvJP, v[50:])
        PMqss, PCqs, Pg, PMqs = map(cf_proj, [tMqss, tCqs, tg, tMqs])
        SMqss, SCqs, Sg, SMqs = map(act_proj, [tMqss, tCqs, tg, tMqs])

        f0 = PMqss + PCqs + Pg
        T, fc = self.robot.get_friction_cones(ncontacts=cdim / 3)

        # QP: minimize norm(z)
        #     s.t. dot(T, f0 + dot(nullspace_proj, z)) \leq fc
        qp_P = cvxopt.matrix(eye(cdim))
        qp_q = cvxopt.matrix(zeros(cdim))
        qp_G = cvxopt.matrix(dot(T, nullspace_proj))
        qp_h = cvxopt.matrix(fc - dot(T, f0))
        qp_z = cvxopt.solvers.qp(qp_P, qp_q, qp_G, qp_h)['x']
        z = array(qp_z).reshape((cdim,))

        # the choice below is arbitrary
        # see the paper for a discussion of possible values
        z0 = zeros(cdim)
        z1 = z
        z2 = zeros(cdim)

        a0 = PMqs + dot(nullspace_proj, z0)
        b0 = PMqss + PCqs + dot(nullspace_proj, z1)
        c0 = Pg + dot(nullspace_proj, z2)

        assert allclose(z1 + z2, z)
        assert (dot(T, b0 + c0) - fc <= 11e-5).all()

        # dot(T, f) <= fc
        a_fric = dot(T, a0)
        b_fric = dot(T, b0)
        c_fric = dot(T, c0) - fc
        a, b, c = map(list, [a_fric, b_fric, c_fric])

        # tau <= +taumax
        a_act, b_act, c_act = {}, {}, {}
        a_act['right'] = SMqs - act_proj(dot(JT, a0))
        b_act['right'] = SMqss + SCqs - act_proj(dot(JT, b0))
        c_act['right'] = Sg - act_proj(dot(JT, c0)) - self.taumax

        # tau >= -taumax
        a_act['left'] = -a_act['right']
        b_act['left'] = -b_act['right']
        c_act['left'] = -Sg + act_proj(dot(JT, c0)) - self.taumax

        for hs in ['left', 'right']:
            for dof in self.active_dofs:
                if dof.torque_limit is not None:
                    a.append(a_act[hs][dof.index])
                    b.append(b_act[hs][dof.index])
                    c.append(c_act[hs][dof.index])

        return a, b, c
